"""
Provides the canvases for plots in muonic
"""
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt4agg \
    import FigureCanvasQTAgg as FigureCanvas
try:
    from matplotlib.backends.backend_qt4agg \
        import NavigationToolbar2QTAgg as NavigationToolbar
except ImportError:
    from matplotlib.backends.backend_qt4agg \
        import NavigationToolbar2QT as NavigationToolbar

import numpy as np


class BasePlotCanvas(FigureCanvas):
    """
    Base class for plot canvases

    :param parent: the parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param ymin: minimum y-value
    :type ymin: float
    :param ymax: maximum y-value
    :type ymax: float
    :param xmin: minimum x-value
    :type xmin: float
    :param xmax: maximum x-value
    :type xmax: float
    :param xlabel: label of the x-axis
    :type xlabel: str
    :param ylabel: label of the y-axis
    :type ylabel: str
    :param grid: draw grid
    :type grid: bool
    :param spacing: left and right spacing of the subplots
    :type spacing: tuple
    """
    
    def __init__(self, parent, logger, ymin=0, ymax=10, xmin=0, xmax=10,
                 xlabel="xlabel", ylabel="ylabel", grid=True,
                 spacing=(0.1, 0.9)):
       
        self.logger = logger

        # initialization of the canvas
        self.fig = Figure(facecolor="white", dpi=72)
        self.fig.subplots_adjust(left=spacing[0], right=spacing[1])
        FigureCanvas.__init__(self, self.fig)

        # setup subplot, axis and grid
        self.ax = self.fig.add_subplot(111)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)
        self.ax.set_autoscale_on(False)
        self.ax.grid(grid)
        
        # store the limits for later use
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.xlabel = xlabel
        self.ylabel = ylabel
        
        # force a redraw of the Figure
        self.fig.canvas.draw() 
        self.setParent(parent)

    # FIXME: is this of any use here? remove?
    @staticmethod
    def color(string, color="none"):
        """
        Color string via ANSI escape sequences

        :param string: the string to color
        :type string: str
        :param color: the color of the string, eg.: 'blue', 'green'
        :type color: str
        :returns: str
        """
        colors = {"green": '\033[92m', "yellow": '\033[93m',
                  "red": '\033[91m', "blue": '\033[94m', "none": '\033[0m'}
        return colors[color] + string + colors["none"]   

    def update_plot(self, *args):
        """
        Instructions to update the plot. Needs to be implemented in subclasses.

        :Returns: None
        """
        raise NotImplementedError("implement this method")


class BaseHistogramCanvas(BasePlotCanvas):
    """
    A base class for all canvases with a histogram

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param binning: the binning to use for this canvas
    :type binning: list or tuple or numpy.ndarray
    :param hist_color: the color of the histogram
    :type hist_color: str
    :param kwargs: additional keyword arguments
    :param kwargs: dict
    """

    def __init__(self, parent, logger, binning, hist_color="b", **kwargs):
        BasePlotCanvas.__init__(self, parent, logger, **kwargs)

        # setup binning
        self.binning = np.asarray(binning)
        self.bincontent = np.zeros(len(self.binning))
        self.hist_patches = self.ax.hist(np.array([self.binning[0] - 1]),
                                         self.binning, fc=hist_color,
                                         alpha=0.25)[2]
        self.heights = []
        self.dimension = r"$\mu$s"

        # FIXME the current implementation does not know about outliers
        self.underflow = 0
        # FIXME the current implementation does not know about outliers
        self.overflow = 0

        # fixed xrange for histogram
        self.xmin = self.binning[0]
        self.xmax = (self.binning[-1] +
                     (self.binning[:-1] - self.binning[1:])[-1])

    def update_plot(self, data):
        """
        Update the plot

        :param data: the data to plot
        :type data: list of lists
        :return: None
        """
        if not data:
            return

        # avoid memory leak
        self.ax.clear()

        # we have to do some bad hacking here,
        # because the p histogram is rather
        # simple and it is not possible to add two of them...
        # however, since we do not want to run into a memory leak
        # and we also be not dependent on dashi (but maybe
        # sometimes in the future?) we have to do it
        # by manipulating rectangles...

        # we want to find the non-empty bins
        # tmp_hist is compatible with the decay_time hist...
        tmp_hist = self.ax.hist(data, self.binning, fc="b", alpha=0.25)[0]

        for hist_bin in enumerate(tmp_hist):
            if hist_bin[1]:
                self.hist_patches[hist_bin[0]].set_height(
                        self.hist_patches[hist_bin[0]].get_height() +
                        hist_bin[1])

        # we want to get the maximum for the ylims
        # self.heights contains the bincontent!

        self.heights = []
        for patch in self.hist_patches:
            self.heights.append(patch.get_height())

        self.logger.debug("Histogram patch heights %s" % self.heights)
        self.ax.set_ylim(ymax=max(self.heights) * 1.1)
        self.ax.set_ylim(ymin=0)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.ax.set_xlim(xmin=self.xmin, xmax=self.xmax)

        # always get rid of unused stuff
        del tmp_hist

        # some beautification
        self.ax.grid()

        # we now have to pass our new patches
        # to the figure we created..
        self.ax.patches = self.hist_patches
        self.fig.canvas.draw()

    def show_fit(self, bin_centers, bincontent, fitx, decay, p, covar,
                 chisquare, nbins):
        """
        Plot the fit onto the diagram

        :param bin_centers: bin centers
        :param bincontent: bincontents
        :param fitx: the fit
        :type fitx: numpy.ndarray
        :param decay: decay function
        :type decay: function
        :param p: fit parameters
        :type p: list
        :param covar: covariance matrix
        :type covar: matrix
        :param chisquare: chi-squared
        :type chisquare: float
        :param nbins: number of bins
        :type nbins: int
        :returns: None
        """

        # clears a previous fit from the canvas
        self.ax.lines = []
        self.ax.plot(bin_centers, bincontent, "b^", fitx, decay(p, fitx), "b-")

        # FIXME: this seems to crop the histogram
        # self.ax.set_ylim(0,max(bincontent)*1.2)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)

        # compute the errors on the fit, nb that this calculation assumes that
        # scipy.optimize.leastsq was used
        error = []
        for i in range(len(p)):
            try:
                error.append(np.absolute(covar[i][i]) ** 0.5)
            except Exception:
                error.append(0.00)

        perr_leastsq = np.array(error)

        try:
            self.ax.legend(("Data", ("Fit: (%4.2f $\pm$ %4.2f) %s \n" +
                                     " chisq/ndf=%4.2f") %
                            (p[2], perr_leastsq[2], self.dimension,
                             chisquare / (nbins-len(p)))), loc=1)
        except TypeError:
            self.logger.warn("Covariance Matrix is 'None', could " +
                             "not calculate fit error!")
            self.ax.legend(("Data", "Fit: (%4.2f) %s \n chisq/ndf=%4.2f" %
                            (p[2], self.dimension, chisquare/(nbins-len(p)))),
                           loc=1)

        self.fig.canvas.draw()

        
class PulseCanvas(BasePlotCanvas):
    """
    Canvas to display pulses

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    """
    def __init__(self, parent, logger):
        BasePlotCanvas.__init__(self, parent, logger, ymin=0, ymax=1.2,
                                xmin=0, xmax=40, xlabel="Time (ns)",
                                ylabel="ylabel", grid=True)
        self.ax.set_title("Oscilloscope")
        self.ax.yaxis.set_visible(False)

    def update_plot(self, pulses):
        # do a complete redraw of the plot to avoid memory leak!
        self.ax.clear()

        # set specific limits for x and y axes
        self.ax.set_xlim(0, 100)
        self.ax.set_ylim(ymax=1.5)
        self.ax.grid()
        self.ax.set_xlabel('Time (ns)')
        self.ax.yaxis.set_visible(False)
        self.ax.set_title("Oscilloscope")

        # and disable figure-wide auto scale
        self.ax.set_autoscale_on(False)

        # we have only the information that the pulse is over the threshold,
        # besides that we do not have any information about its height
        # TODO: It would be nice to implement the thresholds as scaling factors

        colors = ['b', 'g', 'r', 'c']
        labels = ['c0', 'c1', 'c2', 'c3']

        pulse_height = 1.0
        pulse_max = []

        if pulses is None:
            self.logger.warning("Pulses have no value - " +
                                "channels not connected?")
        else:
            for chan in enumerate(pulses[1:]):
                for pulse in chan[1]:
                    self.ax.plot([pulse[0], pulse[0], pulse[1], pulse[1]],
                                 [0, pulse_height, pulse_height, 0],
                                 colors[chan[0]], label=labels[chan[0]], lw=2)
                    pulse_max.append(pulse[0])
                    pulse_max.append(pulse[1])

            pulse_max = max(pulse_max)*1.2
            # TODO: the trick below does not really work as expected. 
            # if pulse_max < self.ax.get_xlim()[1]:
            #    pulse_max = self.ax.get_xlim()[0]
            self.ax.set_xlim(0, pulse_max)

            try:
                self.ax.legend(loc=1, ncol=5, mode="expand",
                               borderaxespad=0., handlelength=1)
            except Exception as e:
                self.logger.info("An error with the legend occurred: %s" % e)
                self.ax.legend(loc=2)

            self.fig.canvas.draw()
        
        
class ScalarsCanvas(BasePlotCanvas):
    """
    A plot canvas to display scalars

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param max_length: maximum number of values to plot
    :type max_length: int
    """
    DEFAULT_CHANNEL_CONFIG = [True, True, True, True]
    CHANNEL_COLORS = ['y', 'm', 'c', 'b']
    TRIGGER_COLOR = 'g'

    def __init__(self, parent, logger, max_length=40):
        
        BasePlotCanvas.__init__(self, parent, logger, ymin=0, ymax=20,
                                xlabel="Time (s)", ylabel="Rate (1/s)")
        self.show_trigger = True
        self.max_length = max_length
        self.channel_data = [[], [], [], []]
        self.trigger_data = []
        self.time_data = []
        self.time_window = 0
        self.reset()
        
    def reset(self, show_pending=False):
        """
        Reset all cached plot data

        :param show_pending: indicate pending state
        :type show_pending: bool
        :returns: None
        """
        self.ax.clear()
        self.ax.grid()
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.ax.set_xlim((self.xmin, self.xmax))
        self.ax.set_ylim((self.ymin, self.ymax))

        self.channel_data = [[], [], [], []]
        self.trigger_data = []
        self.time_data = []
        self.time_window = 0

        for ch in range(4):
            self.ax.plot(self.time_data, self.channel_data[ch],
                         c=self.CHANNEL_COLORS[ch],
                         label=("ch%d" % ch), lw=3)
        if self.show_trigger:
            self.ax.plot(self.time_data, self.trigger_data, c='g',
                         label='trigger', lw=3)

        if show_pending:
            left, width = .25, .5
            bottom, height = .35, .8
            right = left + width
            top = bottom + height
            self.ax.text(0.5 * (left + right), 0.5 * (bottom + top),
                         'Measuring...', horizontalalignment='center',
                         verticalalignment='center', fontsize=56, color='red',
                         fontweight="heavy", alpha=.8, rotation=30,
                         transform=self.fig.transFigure)

        self.fig.canvas.draw()

    def update_plot(self, data, show_trigger=True,
                    enabled_channels=DEFAULT_CHANNEL_CONFIG):
        """
        Update plot

        :param data: plot data
        :type data: list of lists
        :param show_trigger: show trigger in plot
        :type show_trigger: bool
        :param enabled_channels: enabled channels
        :type enabled_channels: list of bool
        :returne: None
        """
        # do a complete redraw of the plot to avoid memory leak!
        self.ax.clear()
        self.show_trigger = show_trigger

        self.ax.grid()
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)

        self.logger.debug("result : %s" % data)

        # update lines data using the lists with new data
        self.time_window += data[5]
        self.time_data.append(self.time_window)

        for ch in range(4):
            self.channel_data[ch].append(data[ch])
            if enabled_channels[ch]:
                self.ax.plot(self.time_data, self.channel_data[ch],
                             c=self.CHANNEL_COLORS[ch],
                             label=("ch%d" % ch), lw=2, marker='v')

        self.trigger_data.append(data[4])

        if self.show_trigger:
            self.ax.plot(self.time_data, self.trigger_data,
                         c=self.TRIGGER_COLOR,
                         label='trg', lw=2, marker='x')

        try:
            # get count of active cannels
            channels = enabled_channels + [show_trigger]
            active_count = sum(channels)
            
            self.ax.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                           ncol=active_count, mode="expand", borderaxespad=0.,
                           handlelength=2)
        except Exception as e:
            self.logger.info("An error with the legend occurred: %s" % e)
            self.ax.legend(loc=2)

        if len(self.channel_data[0]) > self.max_length:
            for ch in range(4):
                self.channel_data[ch].remove(self.channel_data[ch][0])
            self.trigger_data.remove(self.trigger_data[0])
            self.time_data.remove(self.time_data[0])
        
        ma = max(max(self.channel_data[0]), max(self.channel_data[1]),
                 max(self.channel_data[2]), max(self.channel_data[3]),
                 max(self.trigger_data))
            
        self.ax.set_ylim(0, ma * 1.1)

        # do not set x-range if time_data consists of only one item to
        # avoid matlibplot UserWarning
        if len(self.time_data) > 1:
            self.ax.set_xlim(self.time_data[0], self.time_data[-1])

        self.fig.canvas.draw()


class LifetimeCanvas(BaseHistogramCanvas):
    """
    A simple histogram for the use with mu lifetime
    measurement

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param binning: the binning to use for this canvas
    :type binning: list or tuple or numpy.ndarray
    """
    def __init__(self, parent, logger, binning=(0, 10, 21)):
        BaseHistogramCanvas.__init__(
                self, parent, logger,
                np.linspace(binning[0], binning[1], binning[2]),
                xlabel="Time between Pulses ($\mu$s)", ylabel="Events")

     
class VelocityCanvas(BaseHistogramCanvas):
    """
    A simple histogram for the use with mu velocity measurement

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param binning: the binning to use for this canvas
    :type binning: list or tuple or numpy.ndarray
    """
    def __init__(self, parent, logger, binning=(0., 30, 25)):
        BaseHistogramCanvas.__init__(
                self, parent, logger,
                np.linspace(binning[0], binning[1], binning[2]),
                xmin=0., xmax=30, ymin=0, ymax=2,
                ylabel="Events", xlabel="Flight Time (ns)")
        self.dimension = r"$ns$"


class PulseWidthCanvas(BaseHistogramCanvas):
    """
    A simple histogram for the use with pulse width measurement

    :param parent: parent widget
    :param logger: logger object
    :type logger: logging.Logger
    :param hist_color: the color of the histogram
    :type hist_color: str
    """
    def __init__(self, parent, logger, hist_color="r"):
        BaseHistogramCanvas.__init__(
                self, parent, logger, np.linspace(0., 100, 30),
                hist_color=hist_color, xmin=0., xmax=100, ymin=0, ymax=2,
                ylabel="Events", xlabel="Pulse Width (ns)")
        self.ax.set_title("Pulse widths")

    def update_plot(self, data):
        BaseHistogramCanvas.update_plot(self, data)
        self.ax.set_title("Pulse widths")



"""
Utility functions
"""
from __future__ import print_function
import os
import shutil

from muonic import DATA_PATH

_data_path = DATA_PATH


def setup_data_directory(directory=DATA_PATH):
    """
    Tries to create the data directory if it is not present. Also tries to
    set the appropriate permissions if it is not writeable.

    :param directory: the directory to set up
    :type directory: str
    :returns: None
    """
    if not os.path.isdir(directory):
        os.makedirs(directory, 0o755)
    elif not os.access(directory, os.W_OK):
        os.chmod(directory, 0o755)


def set_data_directory(directory):
    """
    Sets the data directory

    :param directory: the data directory
    :type directory: str
    :returns: None
    """
    global _data_path

    _data_path = directory


def get_data_directory():
    """
    Get the data directory. Acts as a proxy for custom data
    directories later on.

    :returns: str
    """
    global _data_path

    return _data_path


def get_muonic_filename(start_date, measurement_id, suffix):
    """
    Get a filename in the right format. 'measurement_id' may be a single char
    identifying the measurement type. 'suffix' may be the initials of the
    current user.

    :param start_date: the start date of the measurement
    :type start_date: date object
    :param measurement_id: the identifier of the measurement
    :type measurement_id: str
    :param suffix: the suffix
    :type suffix: str
    :returns: str
    """
    return os.path.join(get_data_directory(), "%s_%s_HOURS_%s" %
                        (start_date.strftime('%Y-%m-%d_%H-%M-%S'),
                         measurement_id,
                         suffix))


def rename_muonic_file(duration, file_path):
    """
    Replaces the placeholder 'HOURS' in filename by the elapsed amount of
    hours since measurement started.

    Raises OSError if the file_path does not contain a filename.

    :param duration: the duration of the measurement
    :type duration: date object
    :param file_path: the file path
    :type file_path: str
    :raises: OSError
    :returns: bool
    """
    hours = get_hours_from_duration(duration)

    path, filename = os.path.split(file_path)

    if filename == "":
        raise OSError("filename is empty")

    new_filename = filename.replace("HOURS", str(hours))

    # move file
    shutil.move(file_path, os.path.join(path, new_filename))


def get_hours_from_duration(duration):
    """
    Get the hourse from a given duration

    :param duration: the duration of the measurement
    :type duration: date object
    :returns: float
    """
    return round(duration.seconds / 3600., 2) + duration.days * 24


def format_date(date, fmt="%Y-%m-%d_%H-%M-%S"):
    """
    Format time by the provides format.

    :param date: the date
    :type: date object
    :param fmt: the date format
    :type fmt: str
    :returns: str
    """
    return date.strftime(fmt)


class WrappedFile(object):
    """
    A file wrapper which keeps track of open files.

    Raises ValueError if filename is None.

    :param filename: the filename
    :type filename: str
    :raises: ValueError
    """
    open_files = set()

    def __init__(self, filename):
        if filename is None:
            raise ValueError("filename cannot be of 'NoneType'")

        self._filename = filename
        self._file = None

    def get_filename(self):
        """
        Get the filename

        :returns: str
        """
        return self._filename

    def open(self, mode='w'):
        """
        Open file and track it.

        :param mode: the file mode
        :type mode: str
        :returns: None
        """
        self._file = open(self._filename, mode)
        WrappedFile.open_files.add(self._filename)
        return self

    @property
    def closed(self):
        """
        Returns True if the file is closed, False otherwise.

        :returns: bool
        """
        if self._file is None:
            return True
        return False

    def close(self):
        """
        Close file and un-track it.

        Raises IOError if file is not open.

        :raises: IOError
        :returns: None
        """
        if self._file is None:
            raise IOError("file '%s' is not open" % self._filename)
        self._file.close()
        self._file = None
        WrappedFile.open_files.remove(self._filename)

    def __enter__(self):
        """
        Return file on entering 'with' block

        :returns: file
        """
        return self._file

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Close file on leaving 'with' block.

        Exit the runtime context related to this object. The parameters
        describe the exception that caused the context to be exited. If
        the context was exited without an exception, all three arguments
        will be None.

        :param exc_type: type of the Exception that caused the
        context to be exited
        :param exc_value: value of the Exception
        :param traceback: traceback, if any
        :returns: None
        """
        self.close()

    def __getattr__(self, attr):
        """
        Proxy all other attributes of file.

        Raises IOError if file is not open.

        :param attr: attribute name
        :type attr: str
        :raises: IOError
        :returns: mixed
        """
        if self._file is None:
            raise IOError("file '%s' is not open" % self._filename)
        return getattr(self._file, attr)

    def __repr__(self):
        """
        Representation

        :returns: str
        """
        return self._filename

    @staticmethod
    def get_open_files():
        """
        Get all open files.

        :returns: set of str
        """
        return WrappedFile.open_files



"""
The gui of the programm, written with PyQt4
"""
import muonic.gui.helpers
from .application import Application
import muonic.gui.dialogs
import muonic.gui.widgets
import muonic.gui.plot_canvases

__all__ = ["helpers", "Application", "dialogs", "widgets", "plot_canvases"]



"""
Global application settings store
"""
from __future__ import print_function

__all__ = ["update_setting", "have_setting", "get_setting",
           "remove_setting", "update_settings",
           "apply_default_settings", "dump_settings"]

_default_settings = {
    "write_pulses": False,
    "write_daq_status": False,
    "time_window": 5.0,
    "gate_width": 0.0,
    "veto": False,
    "veto_ch0": False,
    "veto_ch1": False,
    "veto_ch2": False,
    "active_ch0": True,
    "active_ch1": True,
    "active_ch2": True,
    "active_ch3": True,
    "coincidence0": True,
    "coincidence1": False,
    "coincidence2": False,
    "coincidence3": False,
    "threshold_ch0": 300,
    "threshold_ch1": 300,
    "threshold_ch2": 300,
    "threshold_ch3": 300
}

_settings = dict()


def update_setting(key, value):
    """
    Update value for settings key.

    Raises KeyError if key is None.

    :param key: settings key
    :type key: str
    :param value: setting value
    :type value: object
    :raises: KeyError
    :returns: None
    """
    global _settings

    if key is None:
        raise KeyError("key must not be of 'None-Type'")

    _settings[key] = value


def have_setting(key):
    """
    Returns true if settings key exists, False otherwise.

    :param key: settings key
    :type key: str
    :returns: bool
    """
    return key in _settings


def get_setting(key, default=None):
    """
    Retrieves the settings value for given key.

    :param key: settings key
    :type key: str
    :param default: the default value if setting is not found
    :type default: mixed
    :returns: object
    """
    return _settings.get(key, default)


def remove_setting(key):
    """
    Remove setting with key.

    :param key: settings key
    :type key: str
    :returns: removed object
    """
    return _settings.pop(key)


def clear_settings():
    """
    Clears the settings store

    :returns: None
    """
    _settings.clear()


def update_settings(settings, clear=False):
    """
    Add settings from dict.

    :param settings: settings dictionary
    :type settings: dict
    :param clear: clear settings store before updating settings
    :type clear: bool
    :returns: None
    """
    if settings is None:
        return
    if isinstance(settings, dict):
        if clear:
            clear_settings()
        for key, value in list(settings.items()):
            update_setting(key, value)
    else:
        raise TypeError("argument has to be a dict")


def apply_default_settings(clear=False):
    """
    Apply default settings. If 'clear' is False settings keys
    different from the default settings will retain.

    :param clear: clear settings store before applying default settings
    :type clear: bool
    :returns: None
    """
    update_settings(_default_settings, clear)


def dump_settings():
    """
    Prints the current settings.

    :returns: None
    """
    for key, value in sorted(_settings.items()):
        print("%-20s = %s" % (key, value))



"""
Provides the public interfaces to read from and send to a DAQ card
"""

from __future__ import print_function
import abc
from future.utils import with_metaclass
import logging
import multiprocessing as mp
import re
import queue

try:
    import zmq
except ImportError:
    # DAQMissingDependencyError will be raised when trying to use zmq
    pass

from muonic.daq import DAQIOError, DAQMissingDependencyError
from muonic.daq import DAQSimulationConnection, DAQConnection


class BaseDAQProvider(with_metaclass(abc.ABCMeta, object)):
    """
    Base class defining the public API and helpers for the
    DAQ provider implementations

    :param logger: logger object
    :type logger: logging.Logger
    """

    LINE_PATTERN = re.compile("^[a-zA-Z0-9+-.,:()=$/#?!%_@*|~' ]*[\n\r]*$")

    def __init__(self, logger=None):
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

    @abc.abstractmethod
    def get(self, *args):
        """
        Get something from the DAQ.

        :param args: queue arguments
        :type args: list
        :returns: str or None
        """
        return

    @abc.abstractmethod
    def put(self, *args):
        """
        Send information to the DAQ.

        :param args: queue arguments
        :type args: list
        :returns: None
        """
        return

    @abc.abstractmethod
    def data_available(self):
        """
        Tests if data is available from the DAQ.

        :returns: int or bool
        """
        return

    def _validate_line(self, line):
        """
        Validate line against pattern. Returns None it the provided line is
        invalid or the line if it is valid.

        :param line: line to validate
        :type line: str
        :returns: str or None
        """
        if self.LINE_PATTERN.match(line) is None:
            # Do something more sensible here, like stopping the DAQ then
            # wait until service is restarted?
            self.logger.warning("Got garbage from the DAQ: %s" %
                                line.rstrip('\r\n'))
            return None
        return line


class DAQProvider(BaseDAQProvider):
    """
    DAQProvider

    :param logger: logger object
    :type logger: logging.Logger
    :param sim: enables DAQ simulation if set to True
    :type sim: bool
    """

    def __init__(self, logger=None, sim=False):
        BaseDAQProvider.__init__(self, logger)
        self.out_queue = mp.Queue()
        self.in_queue = mp.Queue()

        if sim:
            self.daq = DAQSimulationConnection(self.in_queue, self.out_queue,
                                               self.logger)
        else:
            self.daq = DAQConnection(self.in_queue, self.out_queue,
                                     self.logger)
        
        # Set up the thread to do asynchronous I/O. More can be made if
        # necessary. Set daemon flag so that the threads finish when the main
        # app finishes
        self.read_thread = mp.Process(target=self.daq.read, name="pREADER")
        self.read_thread.daemon = True
        self.read_thread.start()

        if not sim:
            self.write_thread = mp.Process(target=self.daq.write,
                                           name="pWRITER")
            self.write_thread.daemon = True
            self.write_thread.start()
        
    def get(self, *args):
        """
        Get something from the DAQ.

        Raises DAQIOError if the queue is empty.

        :param args: queue arguments
        :type args: list
        :returns: str or None -- next item from the queue
        :raises: DAQIOError
        """
        try:
            line = self.out_queue.get(*args)
        except queue.Empty:
            raise DAQIOError("Queue is empty")

        return self._validate_line(line)

    def put(self, *args):
        """
        Send information to the DAQ.

        :param args: queue arguments
        :type args: list
        :returns: None
        """
        self.in_queue.put(*args)

    def data_available(self):
        """
        Tests if data is available from the DAQ.

        :returns: int or bool
        """
        try:
            size = self.out_queue.qsize()
        except NotImplementedError:
            self.logger.debug("Running Mac version of muonic.")
            size = not self.out_queue.empty()
        return size


class DAQClient(BaseDAQProvider):
    """
    DAQClient

    Raises DAQMissingDependencyError if zmq is not installed.

    :param address: address to connect to
    :type address: str
    :param port: TCP port to connect to
    :type port: int
    :param logger: logger object
    :type logger: logging.Logger
    :raises: DAQMissingDependencyError
    """
    
    def __init__(self, address='127.0.0.1', port=5556, logger=None):
        BaseDAQProvider.__init__(self, logger)
        try:
            self.socket = zmq.Context().socket(zmq.PAIR)
            self.socket.connect("tcp://%s:%d" % (address, port))
        except NameError:
            raise DAQMissingDependencyError("no zmq installed...")

    def get(self, *args):
        """
        Get something from the DAQ.

        Raises DAQIOError if the queue is empty.

        :param args: queue arguments
        :type args: list
        :returns: str or None -- next line read from socket
        :raises: DAQIOError
        """
        try:
            line = self.socket.recv_string()
        except Exception:
            raise DAQIOError("Socket error")
        
        return self._validate_line(line)

    def put(self, *args):
        """
        Send information to the DAQ.

        :param args: queue arguments
        :type args: list
        :returns: None
        """
        self.socket.send_string(*args)

    def data_available(self):
        """
        Tests if data is available from the DAQ.

        :returns: int or bool
        """
        return self.socket.poll(200)



"""
Provide the dialog fields for user interaction
"""

from os import path

from PyQt4 import QtCore, QtGui


class BaseDialog(QtGui.QDialog):
    """
    Abstract base class for all dialogs

    :param window_title: the title of the window
    :type window_title: str
    """
    DEFAULT_ITEM_LABELS = ["Chan0", "Chan1", "Chan2", "Chan3"]
    
    def __init__(self, window_title):
        QtGui.QDialog.__init__(self)
        self.setWindowTitle(window_title)
        self.setModal(True)

    def get_widget_value(self, object_name):
        """
        Get the value of a widget by its object name.

        :param object_name: the object name of the widget
        :type: str
        :return: mixed
        """
        widget = self.findChild(QtGui.QWidget, object_name)

        if widget is not None:
            if (isinstance(widget, QtGui.QRadioButton) or
                    isinstance(widget, QtGui.QCheckBox) or
                    isinstance(widget, QtGui.QGroupBox)):
                return widget.isChecked()
            elif (isinstance(widget, QtGui.QSpinBox) or
                    isinstance(widget, QtGui.QDoubleSpinBox)):
                return widget.value()
            elif isinstance(widget, QtGui.QLineEdit):
                return widget.text()
        return None
    
    def button_box(self, left=80, top=900):
        """
        Create a custom button for cancel/apply.

        :param left: left offset
        :type left: int
        :param top: top offset
        :type top: int
        """
        box = QtGui.QDialogButtonBox(self)
        box.setGeometry(QtCore.QRect(left, top, 300, 32))
        box.setOrientation(QtCore.Qt.Horizontal)
        box.setStandardButtons(QtGui.QDialogButtonBox.Cancel |
                               QtGui.QDialogButtonBox.Ok)

        QtCore.QObject.connect(box, QtCore.SIGNAL('accepted()'), self.accept)
        QtCore.QObject.connect(box, QtCore.SIGNAL('rejected()'), self.reject)

        return box

    def choice_group(self, object_name="object", label="label",
                     checkable=False, checked=False, radio=False,
                     checked_items=None, item_labels=None, left=20):
        """
        Create a group of choices.

        :param object_name: the base object name for the checkbox group.
            checkboxes get named 'object_name_%d' where %d is
            the checkboxes' index
        :type object_name: str
        :param label: the label of the checkbox group
        :type label: str
        :param checkable: determined if the group box itself is checkable
        :type checkable: bool
        :param checked: initial state of the checkable group box
        :type checked: bool
        :param radio: create radio buttons instead of check boxes
        :type radio: bool
        :param checked_items: items that are checked initially
        :type checked_items: list of int
        :param item_labels: labels of the checkboxes/radio buttons
        :type item_labels: list
        :param left: left offset
        :type left: int
        """
        if item_labels is None:
            item_labels = self.DEFAULT_ITEM_LABELS

        layout = QtGui.QVBoxLayout()

        layout.addStretch(1)

        group_box = QtGui.QGroupBox(label)
        group_box.setCheckable(checkable)
        group_box.setChecked(checked)
        group_box.setObjectName(object_name)
        group_box.setLayout(layout)

        for index, label in enumerate(item_labels):
            if radio:
                check_box = QtGui.QRadioButton(self)
            else:
                check_box = QtGui.QCheckBox(self)
            check_box.setGeometry(QtCore.QRect(left, 40 + index * 40, 119, 28))
            check_box.setObjectName("%s_%d" % (object_name, index))
            check_box.setText(label)
 
            if checked_items is not None and index in checked_items:
                check_box.setChecked(True)

            layout.addWidget(check_box)

        return group_box


class DecayConfigDialog(BaseDialog):
    """
    Settings for the muondecay
    """
    
    def __init__(self):
        BaseDialog.__init__(self, "Muon Decay Configuration")

        layout = QtGui.QGridLayout(self)

        layout.addWidget(self.choice_group(radio=True, label="Single Pulse",
                                           object_name="single_checkbox",
                                           checked_items=[1], left=20), 0, 0)
        layout.addWidget(self.choice_group(radio=True, label="Double Pulse",
                                           object_name="double_checkbox",
                                           checked_items=[2], left=180), 0, 1)
        layout.addWidget(self.choice_group(radio=True,
                                           label="Software Veto Channel",
                                           object_name="veto_checkbox",
                                           checked_items=[3], left=300), 0, 2)

        min_time_spinbox = QtGui.QSpinBox()
        min_time_spinbox.setObjectName("min_pulse_time")
        min_time_spinbox.setMaximum(2000)
        min_time_spinbox.setValue(400)
        min_time_spinbox.setToolTip("Reject events where the double pulses " +
                                    "are too close together")

        layout.addWidget(min_time_spinbox, 2, 0)
        layout.addWidget(QtGui.QLabel("Minimum time between\n" +
                                      "two pulses (in ns)"), 2, 1)

        pulse_width_group_box = QtGui.QGroupBox("Set conditions on " +
                                                "pulse width")
        pulse_width_group_box.setCheckable(True)
        pulse_width_group_box.setChecked(False)
        pulse_width_group_box.setObjectName("set_pulse_width_conditions")

        pulse_width_layout = QtGui.QGridLayout(pulse_width_group_box)

        pulse_width_items = [
            {
                'object_name': 'min_single_pulse_width',
                'label': 'Min mu pulse width',
                'tooltip': 'Define a MINIMUM width for the MUON pulse',
                'value': 10
            }, {
                'object_name': 'max_single_pulse_width',
                'label': 'Max mu pulse width',
                'tooltip': 'Define a MAXIMUM width for the MUON pulse',
                'value': 300
            }, {
                'object_name': 'min_double_pulse_width',
                'label': 'Min e pulse width',
                'tooltip': 'Define a MINIMUM width for the ELECTRON pulse',
                'value': 5
            }, {
                'object_name': 'max_double_pulse_width',
                'label': 'Max e pulse width',
                'tooltip': 'Define a MINIMUM width for the ELECTRON pulse',
                'value': 300
            }
        ]

        for i, item in enumerate(pulse_width_items):
            spinbox = QtGui.QSpinBox()
            spinbox.setObjectName(item['object_name'])
            spinbox.setSuffix(' ns')
            spinbox.setValue(item['value'])
            spinbox.setToolTip(item['tooltip'])
            spinbox.setMaximum(11000)
            pulse_width_layout.addWidget(spinbox, i, 0)
            pulse_width_layout.addWidget(QtGui.QLabel(item['label']), i, 1)

        layout.addWidget(pulse_width_group_box, 3, 0, 1, 3)
        layout.addWidget(self.button_box(left=200), 4, 2)

        self.show()


class FitRangeConfigDialog(BaseDialog):
    """
    Dialog to configure the fit range

    :param upper_lim: upper limits for the fit
    :type: tuple of float
    :param lower_lim: lower limits for the fit
    :type: tuple of float
    :param dimension: suffix
    :type dimension: str
    """

    def __init__(self, upper_lim=None, lower_lim=None, dimension=''):
        BaseDialog.__init__(self, "Fit Range Configuration")

        layout = QtGui.QGridLayout(self)

        lower = QtGui.QDoubleSpinBox()
        lower.setDecimals(2)
        lower.setSingleStep(0.01)
        lower.setObjectName("lower_limit")
        lower.setSuffix(' %s' % dimension)

        if lower_lim:
            lower.setMinimum(lower_lim[0])
            lower.setMaximum(lower_lim[1])
            lower.setValue(lower_lim[2])

        layout.addWidget(QtGui.QLabel("Lower limit for the fit range: "), 0, 0)
        layout.addWidget(lower, 0, 1)

        upper = QtGui.QDoubleSpinBox()
        upper.setDecimals(2)
        upper.setSingleStep(0.01)
        upper.setObjectName("upper_limit")
        upper.setSuffix(' %s' % dimension)

        if upper_lim:
            upper.setMinimum(upper_lim[0])
            upper.setMaximum(upper_lim[1])
            upper.setValue(upper_lim[2])

        layout.addWidget(QtGui.QLabel("Upper limit for the fit range: "), 1, 0)
        layout.addWidget(upper, 1, 1)
        layout.addWidget(self.button_box(left=200), 2, 0, 2, 0)

        self.show()


class VelocityConfigDialog(BaseDialog):
    """
    Dialog to configure the muon velocity
    """

    def __init__(self):
        BaseDialog.__init__(self, "Muon Velocity Configuration")

        layout = QtGui.QGridLayout(self)
        layout.addWidget(self.choice_group(radio=True, label="Upper Channel",
                                           object_name="upper_checkbox",
                                           checked_items=[0], left=20), 0, 0)
        layout.addWidget(self.choice_group(radio=True, label="Lower Channel",
                                           object_name="lower_checkbox",
                                           checked_items=[1], left=180), 0, 1)
        layout.addWidget(self.button_box(left=200), 1, 1)

        self.show()


class ThresholdDialog(BaseDialog):
    """
    Dialog to adjust the thresholds

    :param thresholds: the threshold for the four channels
    :type thresholds: list of int
    """

    def __init__(self, thresholds):
        BaseDialog.__init__(self, "Thresholds")

        layout = QtGui.QVBoxLayout(self)

        for channel, threshold in enumerate(thresholds):
            spinbox = QtGui.QSpinBox()
            spinbox.setMaximum(1000)
            spinbox.setObjectName("threshold_ch_%d" % channel)
            spinbox.setValue(int(threshold))
            spinbox.setSuffix(' mV')
            layout.addWidget(QtGui.QLabel("Channel %d" % channel))
            layout.addWidget(spinbox)
                        
        layout.addWidget(self.button_box(left=0))

        self.show()


class ConfigDialog(BaseDialog):
    """
    Dialog to set the configuration

    :param channel_states: activation states of the channels
    :type channel_states: list of bool
    :param coincidence_states: coincidence states
    :type coincidence_states: list of bool
    :param veto_enabled: enable veto group
    :type veto_enabled: bool
    :param channel_veto_states: channel veto
    :type channel_veto_states: list of bool
    """
    DEFAULT_CHANNEL_STATES = [True] * 4
    DEFAULT_COINCIDENCE_STATES = [True] + [False] * 3
    DEFAULT_CHANNEL_VETO_STATES = [False] * 3
    
    def __init__(self, channel_states=DEFAULT_CHANNEL_STATES,
                 coincidence_states=DEFAULT_COINCIDENCE_STATES,
                 veto_enabled=False,
                 channel_veto_states=DEFAULT_CHANNEL_VETO_STATES):
        BaseDialog.__init__(self, "Channel Configuration")

        checked_channels = []
        checked_coincidences = []
        checked_channel_vetos = []

        for _channel, checked in enumerate(channel_states):
            if checked:
                checked_channels.append(_channel)

        for _coincidence, checked in enumerate(coincidence_states):
            if checked:
                checked_coincidences.append(_coincidence)

        for _veto, checked in enumerate(channel_veto_states):
            if checked:
                checked_channel_vetos.append(_veto)

        layout = QtGui.QGridLayout(self)
        layout.addWidget(
                self.choice_group(label="Select Channel",
                                  object_name="channel_checkbox",
                                  checked_items=checked_channels,
                                  left=300), 0, 0)
        layout.addWidget(
                self.choice_group(radio=True, label="Trigger Condition",
                                  object_name="coincidence_checkbox",
                                  checked_items=checked_coincidences,
                                  item_labels=["Single", "Twofold",
                                               "Threefold", "Fourfold"],
                                  left=20), 0, 1)
        layout.addWidget(
                self.choice_group(radio=True, label="Veto",
                                  checkable=True, checked=veto_enabled,
                                  object_name="veto_checkbox",


                                  checked_items=checked_channel_vetos,
                                  item_labels=["Chan1", "Chan2", "Chan3"],
                                  left=180), 0, 2)
        layout.addWidget(self.button_box(left=30, top=300), 1, 2, 1, 2)

        self.show()


class AdvancedDialog(BaseDialog):
    """
    Dialog to set advanced configuration options.

    :param gate_width: the gate width of the time window
    :type gate_width: float
    :param time_window: the readout interval
    :type time_window: float
    :param write_daq_status: write DAQ status to raw file
    :type write_daq_status: bool
    """
    
    def __init__(self, gate_width=100, time_window=5.0,
                 write_daq_status=False):
        BaseDialog.__init__(self, "Advanced Configurations")

        layout = QtGui.QGridLayout(self)

        gate_width_box = QtGui.QSpinBox()
        gate_width_box.setSuffix(' ns')
        gate_width_box.setObjectName("gate_width")
        gate_width_box.setMaximum(159990)
        gate_width_box.setSingleStep(10)
        gate_width_box.setMinimum(10)
        gate_width_box.setValue(gate_width)
        gate_width_box.setToolTip("Define a gate width, which is the " +
                                  "time window opened by a trigger")

        layout.addWidget(QtGui.QLabel("Gate width time window " +
                                      "(default: 100 ns): "), 0, 0)
        layout.addWidget(gate_width_box, 0, 1)

        time_window_box = QtGui.QDoubleSpinBox()
        time_window_box.setDecimals(1)
        time_window_box.setSingleStep(0.1)
        time_window_box.setSuffix(' s')
        time_window_box.setObjectName("time_window")
        time_window_box.setMaximum(1000)
        time_window_box.setMinimum(0.01)
        time_window_box.setValue(time_window)
        time_window_box.setToolTip("Define an interval for calculating " +
                                   "and refreshing the rates.")

        layout.addWidget(QtGui.QLabel("Readout Interval " +
                                      "(default: 5 s): "), 1, 0)
        layout.addWidget(time_window_box, 1, 1)

        write_status_checkbox = QtGui.QCheckBox()
        write_status_checkbox.setObjectName("write_daq_status")
        write_status_checkbox.setChecked(write_daq_status)
        write_status_checkbox.setToolTip("Write DAQ status lines to RAW " +
                                         "file, same as option -n.")

        layout.addWidget(QtGui.QLabel("Write DAQ status lines to RAW file: "),
                         2, 0)
        layout.addWidget(write_status_checkbox, 2, 1)
        layout.addWidget(self.button_box(left=30, top=300), 3, 0, 1, 2)

        self.show()


class HelpDialog(BaseDialog):
    """
    Help Dialog for the DAQ commands
    """

    def __init__(self):
        BaseDialog.__init__(self, "DAQ Commands")
        self.resize(600, 480)

        help_file = path.join(path.dirname(__file__), 'daq_commands_help.txt')

        with open(help_file, 'r') as f:
            help_text = f.read()

        text_box = QtGui.QPlainTextEdit(help_text)
        text_box.setReadOnly(True)

        button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok)
        QtCore.QObject.connect(button_box, QtCore.SIGNAL('accepted()'),
                               self.accept)

        layout = QtGui.QVBoxLayout(self)
        layout.addWidget(text_box)
        layout.addWidget(button_box)

        self.show()

if __name__ == "__main__":
    import sys

    app = QtGui.QApplication(sys.argv)
    config_dialog = ConfigDialog()
    decay_dialog = DecayConfigDialog()
    threshold_dialog = ThresholdDialog([42, 42, 42, 42])
    velocity_dialog = VelocityConfigDialog()
    sys.exit(app.exec_())



#!/usr/bin/env python
from __future__ import print_function
import os
import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    import PyQt4
except ImportError:
    print("missing dependency: before installing muonic you have to " +
          "install python-qt4 via your distributions' package manager")
    sys.exit(1)

import muonic


def read_file(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(name=muonic.__name__,
      version=muonic.__version__,
      author=muonic.__author__,
      author_email=muonic.__author_email__,
      description=muonic.__description__,
      long_description=read_file("README.md"),
      license=muonic.__license__,
      keywords=["QNET", "QuarkNET", "Fermilab", "DESY", "DAQ"],
      url=muonic.__source_location__,
      download_url=muonic.__download_url__,
      install_requires=["future", "matplotlib", "numpy", "pyserial", "scipy"],
      platforms=["Ubuntu 12.04"],
      scripts=["bin/muonic", "bin/which_tty_daq"],
      packages=["muonic", "muonic.analysis", "muonic.daq",
                "muonic.gui", "muonic.util"],
      package_data={"muonic": ["daq/simdaq.txt", "gui/daq_commands_help.txt"]},
      classifiers=[
          "License :: OSI Approved :: GNU General Public License v3 or " +
          "later (GPLv3+)",
          "Development Status :: 4 - Beta",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Education",
          "Intended Audience :: Developers",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering :: Physics"
      ])
